# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import asyncio
import gc
import logging
import multiprocessing as mp
import os
import signal
import socket
import ssl
import sys

from importlib import import_module
from io import TextIOWrapper
from shutil import get_terminal_size

from .managers import ProcessManager
from .routes import Routes
from .utils import (
    file_signature, getoptions, log_date, memory_usage, print_logo, server_date
)
from .lib.connections import KeepAliveConnections
from .lib.contexts import WorkerContext
from .lib.executors import MultiThreadExecutor
from .lib.locks import ServerLock

__version__ = '0.5.1'


class Tremolo:
    def __init__(self, name=None):
        self.logger = None if name is None else logging.getLogger(name)
        self.context = WorkerContext()
        self.manager = ProcessManager()
        self.routes = Routes()
        self.middlewares = {
            (): {
                'request': [],
                'response': []
            }
        }
        self.hooks = {
            'worker_start': [],
            'worker_stop': [],
            'connect': [],
            'close': []
        }
        self.ports = {}

    @property
    def loop(self):
        return asyncio.get_event_loop()

    def add_task(self, task):
        self.context.tasks.add(task)
        task.add_done_callback(self.context.tasks.discard)

    def create_task(self, coro):
        task = self.loop.create_task(coro)
        self.add_task(task)

        return task

    def route(self, path, **options):
        def decorator(func):
            self.add_route(func, path, **options)
            return func

        return decorator

    def error(self, code, func=None):
        i = code - 400

        if not 0 <= i < len(self.routes[0]):
            raise ValueError(f'{code} is not in the supported range: 400-511')

        def decorator(func):
            self.routes[0][i] = (code, func, getoptions(func), {})
            return func

        if callable(func):
            return decorator(func)

        return decorator

    def hook(self, name, *args, priority=999):
        def decorator(func):
            self.add_hook(func, name, priority)
            return func

        if len(args) == 1 and callable(args[0]):
            return decorator(args[0])

        return decorator

    def on_worker_start(self, *args, **kwargs):
        return self.hook('worker_start', *args, **kwargs)

    def on_worker_stop(self, *args, **kwargs):
        return self.hook('worker_stop', *args, **kwargs)

    def on_connect(self, *args, **kwargs):
        return self.hook('connect', *args, **kwargs)

    def on_close(self, *args, **kwargs):
        return self.hook('close', *args, **kwargs)

    def middleware(self, name, *args, priority=999):
        def decorator(func):
            self.add_middleware(func, name, priority, kwargs=getoptions(func))
            return func

        if len(args) == 1 and callable(args[0]):
            return decorator(args[0])

        return decorator

    def on_request(self, *args, **kwargs):
        return self.middleware('request', *args, **kwargs)

    def on_response(self, *args, **kwargs):
        return self.middleware('response', *args, **kwargs)

    def add_route(self, func, path='/', **options):
        if callable(func) and not isinstance(func, type):
            self.routes.add(func, path, getoptions(func))
        else:  # a class-based view
            name = getattr(func, '__module__', None) or func.__name__

            for attr_name in dir(func):
                method = getattr(func, attr_name)

                if attr_name.startswith('_') or not callable(method):
                    continue

                if method.__module__ == name and not isinstance(method, type):
                    self.routes.add(
                        method, path, dict(getoptions(method), self=func),
                        **options
                    )

    def add_hook(self, func, name='worker_start', priority=999):
        if name not in self.hooks:
            raise ValueError('%s is not one of the: %s' %
                             (name, ', '.join(self.hooks)))

        self.hooks[name].append((priority, func))
        self.hooks[name].sort(key=lambda item: item[0],
                              reverse=name in ('worker_stop', 'close'))

    def add_middleware(self, func, name='request', priority=999, *,
                       kwargs=None, prefix=()):
        if name not in self.middlewares[()]:
            raise ValueError('%s is not one of the: %s' %
                             (name, ', '.join(self.middlewares[()])))

        if prefix not in self.middlewares:
            self.middlewares[prefix] = {
                'request': [],
                'response': []
            }

        self.middlewares[prefix][name].append(
            (priority, func, kwargs or getoptions(func))
        )
        self.middlewares[prefix][name].sort(key=lambda item: item[0],
                                            reverse=name == 'response')

    def listen(self, port, host=None, **options):
        if not isinstance(port, int):
            # assume it's a UNIX socket path
            host = port
            port = None

        return self.ports.setdefault((host, port), options) is options

    def mount(self, prefix, app):
        if not prefix.startswith('/') or len(prefix) > 255:
            raise ValueError('prefix must start with "/" and <=255 in length')

        if app is self or not isinstance(app, self.__class__):
            raise ValueError('invalid app')

        prefix = prefix.rstrip('/').encode('latin-1')

        while app.routes:
            _, routes = app.routes.popitem()

            for pattern, func, kwargs, options in routes:
                if isinstance(pattern, bytes):
                    pattern = pattern.lstrip(b'^')

                    if pattern.startswith(b'/'):
                        pattern = b'^' + prefix + pattern
                    else:
                        pattern = b'^' + prefix + b'/.*' + pattern

                    self.routes[-1].append((pattern, func, kwargs, options))

        parts = tuple(part for part in prefix.split(b'/') if part)

        while app.middlewares:
            p, middlewares = app.middlewares.popitem()

            for name in middlewares:
                for middleware in middlewares[name]:
                    self.add_middleware(middleware[1], name, middleware[0],
                                        kwargs=middleware[2], prefix=parts + p)

        while app.hooks:
            name, hooks = app.hooks.popitem()

            for priority, func in hooks:
                self.add_hook(func, name, priority)

        while app.ports:
            self.ports.setdefault(*app.ports.popitem())

    async def serve(self, host, port, *, sock=None, backlog=100, **kwargs):
        options = self.context.options
        options.update(kwargs)

        options.setdefault('debug', False)
        options.setdefault('log_level',
                           'DEBUG' if options['debug'] else 'INFO')
        options.setdefault('app', None)
        options.setdefault('app_dir', os.getcwd())
        options.setdefault('shutdown_timeout', 30)
        options.setdefault('server_name', 'Tremolo')

        if self.logger is None:
            self.logger = logging.getLogger(mp.current_process().name)

        self.logger.setLevel(options['log_level'].upper())

        for handler in self.logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                break
        else:
            handler = logging.StreamHandler()
            formatter = logging.Formatter(
                options.get(
                    'log_fmt',
                    '[%(asctime)s] %(module)s: %(levelname)s: %(message)s')
            )

            handler.setFormatter(formatter)
            self.logger.addHandler(handler)

        if sock is None:
            sock = self.create_sock(host, port)

        sock.listen(backlog)

        if 'ssl' in options and isinstance(options['ssl'] or None, dict):
            ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_context.load_cert_chain(
                certfile=options['ssl'].get('cert', ''),
                keyfile=options['ssl'].get('key'),
                password=options['ssl'].get('password')
            )
        else:
            ssl_context = None

        executor = MultiThreadExecutor(size=options.get('thread_pool_size', 5))
        await executor.start(loop=self.loop)

        if '_locks' in options:
            lock = ServerLock(options['_locks'], executor, loop=self.loop)
        else:
            lock = None

        connections = KeepAliveConnections(
            maxlen=options.get('keepalive_connections', 512)
        )
        self.context.update(connections=connections, executor=executor)
        options['state'] = {}

        for _, func in self.hooks['worker_start']:
            try:
                coro = func(app=self,
                            globals=self.context,
                            context=self.context,
                            loop=self.loop,
                            logger=self.logger)
            except TypeError:
                coro = func(app=self)

            if await coro:
                break

        if options['app'] is None:
            from .http_server import HTTPServer as Server

            self.routes.compile(executor=executor)
        else:
            from .asgi_lifespan import ASGILifespan
            from .asgi_server import ASGIServer as Server

            # 'module:app'               -> 'module:app'   (dir: os.getcwd())
            # '/path/to/module.py'       -> 'module:app'   (dir: '/path/to')
            # '/path/to/module.py:myapp' -> 'module:myapp' (dir: '/path/to')

            if isinstance(options['app'], str):
                if options['app'].find(':',
                                       options['app'].find(':\\') + 1) == -1:
                    options['app'] += ':app'

                path, attr_name = options['app'].rsplit(':', 1)
                options['app_dir'], basename = os.path.split(
                    os.path.abspath(path))
                module_name = os.path.splitext(basename)[0]

                if options['app_dir'] == '':
                    options['app_dir'] = os.getcwd()

                sys.path.insert(0, options['app_dir'])
                options['app'] = getattr(import_module(module_name), attr_name)

            print(log_date(), end=' ')
            print(
                'Starting %s as an ASGI server for:' % options['server_name'],
                end=' '
            )
            print(
                getattr(options['app'], '__name__',
                        options['app'].__class__.__name__)
            )

            if options['server_name'] == 'Tremolo':
                options['server_name'] += ' (ASGI)'

            self.context.lifespan = ASGILifespan(self, options=options)
            self.context.lifespan.startup()
            exc = await self.context.lifespan.exception(
                timeout=options['shutdown_timeout'] / 2
            )

            if exc:
                raise exc

        sockname = sock.getsockname()

        if isinstance(sockname, tuple):
            self.context.info['server'] = sockname[:2]
        else:
            self.context.info['server'] = (sockname, None)

        self.context.info['server_date'] = server_date()
        self.context.info['server_name'] = options[
                                           'server_name'].encode('latin-1')

        options.setdefault('experimental', False)
        options.setdefault('ws', True)
        options.setdefault('ws_max_payload_size', 2 * 1048576)
        options.setdefault('download_rate', 1048576)
        options.setdefault('upload_rate', 1048576)
        options.setdefault('buffer_size', 16384)
        options.setdefault('client_max_body_size', 2 * 1048576)
        options.setdefault('client_max_header_size', 8192)
        options.setdefault('max_queue_size', 128)
        options.setdefault('request_timeout', 30)
        options.setdefault('keepalive_timeout', 30)
        options.setdefault('app_handler_timeout', 120)
        options.setdefault('app_close_timeout', 30)
        options.setdefault('root_path', '')

        server = await self.loop.create_server(
            lambda: Server(app=self, lock=lock, options=options),
            sock=sock, backlog=backlog, ssl=ssl_context
        )

        print(log_date(), end=' ')
        print(
            '%s worker (pid %d) is started at' % (options['server_name'],
                                                  os.getpid()),
            end=' '
        )
        print(self.context.info['server'][0], end=' ')

        if self.context.info['server'][1] is not None:
            print('port %d' % self.context.info['server'][1], end=' ')

        if ssl_context is not None:
            print('(https)', end='')

        print()

        try:
            await self._serve_forever()
        except BaseException as exc:
            self.logger.info('Shutting down: %s', str(exc))

            options['request_timeout'] = 1
            options['keepalive_timeout'] = 0
            options['app_handler_timeout'] = 1
            options['app_close_timeout'] = 1

            while self.context.tasks:
                _, pending = await asyncio.wait(
                    self.context.tasks, timeout=options['shutdown_timeout'] / 2
                )
                for task in pending:
                    task.cancel()

            raise
        finally:
            server.close()

            await server.wait_closed()
            await self._worker_stop()
            await executor.shutdown()

    async def _serve_forever(self):
        limit_memory = self.context.options.get('limit_memory', 0)
        paths = [path for path in sys.path
                 if not self.context.options['app_dir'].startswith(path)]
        modules = {}
        set_threshold = getattr(gc, 'set_threshold', None)

        while True:
            await asyncio.sleep(1)

            if set_threshold:
                set_threshold(0, int(len(self.context.tasks) ** 0.5) + 10)

                if gc.get_count()[1] < gc.get_threshold()[1]:
                    gc.collect(0)
                else:
                    n = gc.collect()

                    if n:
                        self.logger.info('collected %d unreachable objects', n)

            # update server date
            self.context.info['server_date'] = server_date()

            # detect code changes
            if self.context.options.get('reload', False):
                for module in (dict(modules) or sys.modules.values()):
                    module_file = getattr(module, '__file__', None)

                    if module_file is None:
                        continue

                    for path in paths:
                        if module_file.startswith(path):
                            break
                    else:
                        if not os.path.exists(module_file):
                            if module in modules:
                                del modules[module]

                            continue

                        sign = file_signature(module_file)

                        if module in modules:
                            if modules[module] == sign:
                                # file not modified
                                continue

                            modules[module] = sign
                        else:
                            modules[module] = sign
                            continue

                        self.logger.info('reload: %s', module_file)
                        sys.exit(3)

            if limit_memory > 0 and memory_usage() > limit_memory:
                while self.context.tasks:
                    self.context.tasks.pop().cancel()

                self.logger.error('memory limit exceeded')
                sys.exit(1)

    async def _worker_stop(self):
        try:
            if self.context.options['app'] is not None:
                self.context.lifespan.shutdown()
                exc = await self.context.lifespan.exception(
                    timeout=self.context.options['shutdown_timeout'] / 2
                )

                if exc:
                    self.logger.error(exc)
        finally:
            for _, func in reversed(self.hooks['worker_stop']):
                try:
                    coro = func(app=self,
                                globals=self.context,
                                context=self.context,
                                loop=self.loop,
                                logger=self.logger)
                except TypeError:
                    coro = func(app=self)

                if await coro:
                    break

    def _worker(self, host, port, **kwargs):
        if not getattr(sys.stdout, 'line_buffering', False):
            sys.stdout = TextIOWrapper(sys.stdout.buffer, line_buffering=True)

        loop_name = kwargs.get('loop', 'asyncio.')

        if '.' not in loop_name:
            loop_name += '.'

        # 'asyncio', '.', ''
        # 'asyncio', '.', 'SelectorEventLoop'
        module_name, _, class_name = loop_name.rpartition('.')
        module = import_module(module_name or 'asyncio')
        loop = getattr(module, class_name or 'new_event_loop')()
        task = loop.create_task(self.serve(host, port, **kwargs))

        task.add_done_callback(lambda fut: loop.stop())
        signal.signal(signal.SIGINT, lambda signum, frame: task.cancel())
        signal.signal(signal.SIGTERM, lambda signum, frame: task.cancel())

        try:
            loop.run_forever()  # until loop.stop() is called
        except BaseException:
            while self.context.tasks:
                self.context.tasks.pop().cancel()

            task.cancel()
            loop.run_forever()

            raise
        finally:
            loop.close()

            if not task.cancelled():
                exc = task.exception()

                if exc:
                    raise exc

    def create_sock(self, host, port, reuse_port=True):
        try:
            try:
                socket.getaddrinfo(host, None)

                if ':' in host:
                    sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)

                    if host == '::' and hasattr(socket, 'IPPROTO_IPV6'):
                        # on Windows, Python versions below 3.8
                        # don't properly support dual-stack IPv4/6.
                        # https://github.com/python/cpython/issues/73701
                        sock.setsockopt(socket.IPPROTO_IPV6,
                                        socket.IPV6_V6ONLY, 0)
                else:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            except socket.gaierror:
                _host = host
                host = 'localhost'
                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                host = _host
        except AttributeError:
            print('either AF_INET6 or AF_UNIX is not supported')
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        sock.setblocking(False)

        if sock.family.name == 'AF_UNIX':
            if not host.endswith('.sock'):
                host += '.sock'

            for _ in range(2):
                try:
                    sock.bind(host)
                    break
                except OSError:
                    if os.path.exists(host) and os.stat(host).st_size == 0:
                        os.unlink(host)
        else:
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            if reuse_port and hasattr(socket, 'SO_REUSEPORT'):
                sock.setsockopt(
                    socket.SOL_SOCKET,
                    getattr(socket, 'SO_REUSEPORT_LB', socket.SO_REUSEPORT),
                    1
                )

            sock.bind((host, port))

        return sock

    def close_sock(self, sock):
        try:
            sock.shutdown(socket.SHUT_RDWR)
            sock.close()

            if sock.family.name == 'AF_UNIX':
                os.unlink(sock.getsockname())
        except FileNotFoundError:
            pass
        except OSError:
            sock.close()

    def _handle_reload(self, kwargs):
        print('Reloading...')

        if kwargs['app'] is None:
            for module in list(sys.modules.values()):
                module_file = getattr(module, '__file__', None)

                if (module_file and
                        module.__name__ not in ('__main__',
                                                '__mp_main__',
                                                'tremolo') and
                        not module.__name__.startswith('tremolo.') and
                        module_file.startswith(kwargs['app_dir']) and
                        os.path.exists(module_file)):
                    del sys.modules[module.__name__]

            module = import_module(kwargs['module_name'])

            # we need to update/rebind objects like
            # routes, middlewares, etc.
            for attr in module.__dict__.values():
                if isinstance(attr, self.__class__) and attr.routes:
                    self.__dict__.update(attr.__dict__)

    def _process_exit(self, **info):
        args = info['args']
        kwargs = info['kwargs']
        process = info['process']

        if process.exitcode == 3:
            self._handle_reload(kwargs)
        elif process.exitcode != 0:
            print(
                'A worker process died (%d). Restarting...' % process.exitcode
            )

        if process.exitcode == 0:
            print('pid %d terminated' % process.pid)
        else:
            self.manager.spawn(
                self._worker, args=args, kwargs=kwargs,
                name=process.name,
                exit_cb=self._process_exit
            )

    def run(self, host=None, port=0, *, worker_num=1, **kwargs):
        server_name = kwargs.get('server_name', 'Tremolo')
        terminal_width = min(get_terminal_size()[0], 72)

        if server_name == 'Tremolo' and terminal_width > 44:
            print_logo()

        print(
            'Starting %s (tremolo %s, %s %d.%d.%d, %s)' %
            (server_name,
             __version__,
             sys.implementation.name,
             *sys.version_info[:3],
             sys.platform)
        )
        print('-' * terminal_width)

        import __main__

        if 'app' in kwargs and kwargs['app']:
            if not isinstance(kwargs['app'], str):
                if not hasattr(__main__, '__file__'):
                    raise RuntimeError('could not find ASGI app')

                for attr_name, attr in __main__.__dict__.items():
                    if attr == kwargs['app']:
                        break
                else:
                    attr_name = 'app'

                kwargs['app'] = '%s:%s' % (__main__.__file__, attr_name)
        else:
            if not self.routes:
                raise RuntimeError('cannot run this app. mounted somewhere?')

            kwargs['app'] = None

            if hasattr(__main__, '__file__'):
                kwargs['app_dir'], basename = os.path.split(
                    os.path.abspath(__main__.__file__)
                )
                kwargs['module_name'] = os.path.splitext(basename)[0]
            else:
                kwargs['app_dir'] = os.getcwd()
                kwargs['module_name'] = '__main__'

            if kwargs.get('log_level', 'DEBUG').upper() in ('DEBUG', 'INFO'):
                print('Routes:')

                for routes in self.routes.values():
                    for pattern, func, kw, _ in routes:
                        if func is None:
                            continue

                        print(
                            '  %s -> %s(%s)' %
                            (pattern,
                             func.__name__,
                             ', '.join('%s=%s' % item for item in kw.items()))
                        )

                print()

        if host is None:
            if not self.ports:
                raise ValueError(
                    'with host=None, listen() must be called first'
                )

            host = 'localhost'
        else:
            self.listen(port, host=host, **kwargs)

        while worker_num == 0:  # single-process mode
            try:
                self._worker(host, port, **kwargs)
                return
            except SystemExit as exc:
                if exc.code == 3:
                    self._handle_reload(kwargs)
                    continue

                raise

        if worker_num < 0:
            try:
                worker_num = len(os.sched_getaffinity(0))
            except AttributeError:
                worker_num = os.cpu_count() or 1

        locks = [mp.Lock() for _ in range(kwargs.get('thread_pool_size', 5))]
        socks = {}
        print('Options:')

        for (_host, _port), options in self.ports.items():
            if _host is None:
                _host = host

            if _port is None:
                _port = port

            options = {**kwargs, **options}
            print(
                '  run(host=%s, port=%d, worker_num=%d, %s)' %
                (_host,
                 _port,
                 worker_num,
                 ', '.join('%s=%s' % item for item in options.items()))
            )

            args = (_host, _port)

            if options.get('reuse_port', hasattr(socket, 'SO_REUSEPORT')):
                sock = None
            else:
                sock = self.create_sock(_host, _port, reuse_port=False)
                socks[args] = sock

            for _ in range(options.get('worker_num', worker_num)):
                self.manager.spawn(
                    self._worker,
                    args=args,
                    kwargs=dict(options, sock=sock, _locks=locks),
                    exit_cb=self._process_exit
                )

        print('-' * terminal_width)
        print('%s main (pid %d) is running ' % (server_name, os.getpid()))

        try:
            self.manager.wait(timeout=kwargs.get('shutdown_timeout', 30))
        finally:
            for sock in socks.values():
                self.close_sock(sock)



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import asyncio
import queue

from functools import wraps
from inspect import isgeneratorfunction
from threading import Thread


def set_result(fut, result):
    if not fut.done():
        fut.set_result(result)


def set_exception(fut, exc):
    if not fut.done():
        fut.set_exception(exc)


class ThreadExecutor(Thread):
    def __init__(self, context, loop=None, **kwargs):
        super().__init__(**kwargs)

        self.context = context
        self.queue = getattr(queue, 'SimpleQueue', queue.Queue)()
        self.loop = loop or asyncio.get_event_loop()

        self._startup = self.loop.create_future()
        self._shutdown = None

    def start(self):
        super().start()
        return self._startup

    def run(self):
        self._startup.set_result(None)

        while self.loop.is_running():
            try:
                self.loop.call_soon_threadsafe(  # set the last active thread
                    self.context.__setattr__, 'thread', self
                )
                fut, func, args, kwargs = self.queue.get(timeout=1)
            except queue.Empty:
                continue

            if func is None:
                self.loop.call_soon_threadsafe(self.join)
                self.loop.call_soon_threadsafe(set_result, fut, None)
                break

            try:
                self.loop.call_soon_threadsafe(
                    set_result, fut, func(*args, **kwargs)
                )
            except StopIteration:
                # StopIteration interacts badly with generators
                # and cannot be raised into a Future
                self.loop.call_soon_threadsafe(fut.cancel)
            except BaseException as exc:
                self.loop.call_soon_threadsafe(set_exception, fut, exc)

    def submit(self, func, *args, **kwargs):
        if not self.is_alive():
            raise RuntimeError(
                'calling submit() before start() or after shutdown()'
            )

        if isgeneratorfunction(func):
            gen = func(*args, **kwargs)

            @wraps(func)
            async def wrapper():
                while True:
                    fut = self.loop.create_future()
                    self.queue.put_nowait((fut, gen.__next__, (), {}))

                    try:
                        yield await fut
                    except asyncio.CancelledError:
                        break

            return wrapper()

        if callable(func):
            fut = self.loop.create_future()
            self.queue.put_nowait((fut, func, args, kwargs))

            return fut

        raise TypeError(f'{str(func)} is not generator function or callable')

    def shutdown(self):
        if self._shutdown is None:
            self._shutdown = self.loop.create_future()

            if self.is_alive():
                self.queue.put_nowait((self._shutdown, None, None, None))

        if not self.is_alive():
            set_result(self._shutdown, None)

        return self._shutdown


class MultiThreadExecutor:
    def __init__(self, size=5):
        self.size = size
        self.threads = []
        self.thread = None  # points to the last active thread
        self.counter = 1

    def start(self, prefix='MultiThreadExecutor', loop=None, **kwargs):
        aws = []

        while len(self.threads) < self.size:
            self.thread = ThreadExecutor(
                self, name=f'{prefix}-{self.counter}', loop=loop, **kwargs
            )
            self.thread.start()
            self.threads.append(self.thread)
            aws.append(self.thread._startup)

            self.counter += 1

        return asyncio.gather(*aws)

    def submit(self, func, args=(), kwargs={}, name=None):
        if self.size == 0 or len(self.threads) < self.size:
            raise RuntimeError('no threads are running or not ready')

        if name is None:
            thread = self.thread
        else:
            thread = self.threads[name % self.size]

        try:
            return thread.submit(func, *args, **kwargs)
        except RuntimeError:  # dead thread found. attempt self-healing
            self.threads.remove(thread)
            self.start()
            raise

    def shutdown(self):
        self.size = 0
        aws = []

        while self.threads:
            aws.append(self.threads.pop().shutdown())

        return asyncio.gather(*aws)



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import asyncio

from .contexts import RequestContext


class Request:
    __slots__ = ('protocol', 'context', 'body_size', 'timeout')

    def __init__(self, protocol):
        self.protocol = protocol
        self.context = RequestContext()
        self.body_size = 0
        self.timeout = protocol.options['request_timeout']

    @property
    def server(self):
        if self.protocol is None:
            raise RuntimeError('protocol object has been closed')

        return self.protocol

    @property
    def ctx(self):
        return self.context

    @property
    def transport(self):
        return self.server.context.transport

    @property
    def socket(self):
        return self.transport.get_extra_info('socket')

    @property
    def client(self):
        return self.server.context.client

    @property
    def is_secure(self):
        return bool(self.transport.get_extra_info('sslcontext'))

    def clear(self):
        self.body_size = 0

        self.context.clear()
        self.protocol = None  # cut off access to the protocol object

    async def recv(self, timeout=None):
        if timeout is not None:
            self.timeout = timeout

        while self.server.queue:
            try:
                data = await self.server.queue[0].get(self.timeout)
            except asyncio.CancelledError as exc:
                raise TimeoutError(0, 'recv timeout') from exc

            if data is None:  # only occurs on a request with Content-Length
                break

            yield data



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import asyncio


class ServerLock:
    def __init__(self, locks, executor, name=0, timeout=30, loop=None):
        self.locks = locks
        self.name = name % len(locks)
        self._executor = executor
        self._timeout = timeout
        self._loop = loop or asyncio.get_event_loop()

    def __call__(self, name=0, *, timeout=None):
        if timeout is None:
            timeout = self._timeout

        return self.__class__(self.locks,
                              self._executor,
                              name=name,
                              timeout=timeout,
                              loop=self._loop)

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(self, exc_type, exc, tb):
        self.release()

    async def acquire(self, timeout=None):
        if timeout is None:
            timeout = self._timeout

        fut = self._executor.submit(self.locks[self.name].acquire,
                                    args=(True, timeout),
                                    name=self.name)
        timer = self._loop.call_at(self._loop.time() + timeout, fut.cancel)

        try:
            result = await fut
        except asyncio.CancelledError:
            result = False
        finally:
            timer.cancel()

        if not result:
            raise TimeoutError

    def release(self):
        try:
            self.locks[self.name].release()
        except ValueError:
            pass



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto


class Response:
    __slots__ = ('request', '_send_buf')

    def __init__(self, request):
        self.request = request
        self._send_buf = bytearray()

    def close(self):
        if self._send_buf:
            self.send_nowait(self._send_buf[:])
            del self._send_buf[:]

        self.send_nowait(None)

    async def send(self, data, rate=-1,
                   buffer_size=16384, buffer_min_size=None):
        if data is None:
            __class__.close(self)
            return

        if not isinstance(data, (bytes, bytearray)):
            raise TypeError('expected None or bytes-like object')

        self._send_buf.extend(data)

        if buffer_min_size is None or len(self._send_buf) >= buffer_min_size:
            while self._send_buf:
                data = self._send_buf[:buffer_size]

                if await self.request.server.put_to_queue(
                        data, name=1, rate=rate):
                    del self._send_buf[:len(data)]
                else:
                    del self._send_buf[:]
                    self.request.clear()

    def send_nowait(self, data):
        if self.request.server.queue:
            self.request.server.queue[1].put_nowait(data)



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import asyncio

from tremolo.utils import parse_int
from .contexts import ConnectionContext
from .http_exceptions import (
    HTTPException,
    BadRequest,
    ExpectationFailed,
    RequestTimeout
)
from .http_header import HTTPHeader
from .http_request import HTTPRequest
from .queue import Queue


# pylint: disable=E0237,E1101
class HTTPProtocol(asyncio.Protocol):
    __slots__ = ('server', 'queue', 'events', 'handlers', '_recv_buf')

    def __init__(self, app, **kwargs):
        self.app = app
        self.extras = kwargs
        self.loop = app.loop
        self.logger = app.logger
        self.globals = app.context  # a worker-level context
        self.context = ConnectionContext()
        self.lock = kwargs['lock']
        self.fileno = -1
        self.request = None  # current request object

        self.server = self.__dict__.copy()  # all the unslotted objects above
        self.queue = [Queue(), Queue()]  # IN, OUT
        self.events = {}
        self.handlers = set()
        self._recv_buf = bytearray()

    @property
    def options(self):
        return self.extras['options']

    @property
    def transport(self):
        return self.context.transport

    def add_close_callback(self, callback):
        self.context.tasks.add(callback)

    def create_task(self, coro):
        task = self.loop.create_task(coro)

        self.context.tasks.add(task)
        task.add_done_callback(self.task_done)

        return task

    def task_done(self, task):
        self.context.tasks.discard(task)

        if not task.cancelled():
            exc = task.exception()

            if exc:
                self.close(exc)

    def connection_made(self, transport):
        self.context.update(transport=transport)
        self.fileno = transport.get_extra_info('socket').fileno()

        self.events['request'] = self.loop.create_future()

        self.app.add_task(self.create_task(self._send_data()))
        self.app.add_task(self.create_task(
            self.set_timeout(self.events['request'],
                             timeout=self.options['request_timeout'],
                             timeout_cb=self.request_timeout)
        ))

    def is_closing(self):
        return self.transport is None or self.transport.is_closing()

    def close(self, exc=None):
        if self.is_closing():
            return

        if exc:
            if isinstance(exc, HTTPException) and exc.code < 600:
                self.transport.write(
                    b'HTTP/1.0 %d %s\r\nContent-Type: %s\r\n\r\n' %
                    (exc.code,
                     exc.message.encode(exc.encoding),
                     exc.content_type.encode(exc.encoding))
                )
                self.transport.write(str(exc).encode(exc.encoding))

            self.print_exception(exc, 'close')

        try:
            self.transport.write_eof()
        except (OSError, RuntimeError) as exc:
            self.logger.info(exc)
        finally:
            self.add_close_callback(
                self.loop.call_at(self.loop.time() +
                                  self.options['app_close_timeout'],
                                  self.transport.abort).cancel
            )
            self.transport.close()

    def request_timeout(self, timeout):
        raise RequestTimeout('request timeout after %gs' % timeout)

    def keepalive_timeout(self, timeout):
        self.logger.debug('keepalive timeout after %gs', timeout)
        self.close()

    def send_timeout(self, timeout):
        self.logger.debug('send timeout after %gs', timeout)

    async def set_timeout(self, waiter, timeout=30, timeout_cb=None):
        timer = self.loop.call_at(self.loop.time() + timeout, waiter.cancel)

        try:
            return await waiter
        except asyncio.CancelledError:
            if not self.is_closing():
                if callable(timeout_cb):
                    timeout_cb(timeout)

                raise
        finally:
            timer.cancel()

    async def put_to_queue(self, data, name=0, rate=-1):
        if self.queue:
            self.queue[name].put_nowait(data)
            queue_size = self.queue[name].qsize()

            if queue_size <= self.options['max_queue_size']:
                if data and rate > 0 and queue_size > rate / len(data):
                    await asyncio.sleep(queue_size * len(data) / rate)

                return True

            self.logger.error('%d exceeds the value of max_queue_size',
                              queue_size)

        self.close()

    async def request_received(self, request, response):
        raise NotImplementedError

    async def error_received(self, exc, response):
        exc = HTTPException(cause=exc)
        i = exc.code - 400

        if 0 <= i < len(self.app.routes[0]) and not response.headers_sent():
            if self.app.routes[0][i][1] is None:
                i = 100

            _, func, kwargs, _ = self.app.routes[0][i]

            response.headers.clear()
            response.set_status(exc.code, exc.message)
            response.set_content_type(exc.content_type)

            return await func(func=func,
                              kwargs=kwargs,
                              request=response.request,
                              response=response,
                              loop=self.loop,
                              exc=exc)

    def handlers_timeout(self):
        if self.request is None or not self.request.upgraded:
            while self.handlers:
                self.handlers.pop().cancel()
                self.logger.error(
                    'handler timeout '
                    '(app_handler_timeout=%g, app_close_timeout=%g)',
                    self.options['app_handler_timeout'],
                    self.options['app_close_timeout']
                )

    def set_handler_timeout(self, timeout):
        if self.handlers:
            return self.loop.call_at(
                self.loop.time() + timeout, self.handlers_timeout
            )

    def print_exception(self, exc, *args):
        cause = exc
        args += (exc.__class__.__name__,)
        errno = -1

        while isinstance(cause, HTTPException):
            if cause.code < 500:
                errno = 0

            if cause.__cause__ is None:
                break

            cause = cause.__cause__
            args += (cause.__class__.__name__,)

        if cause.args and isinstance(cause.args[0], int):
            errno = cause.args[0]
        elif errno == -1:
            errno = getattr(cause, 'errno', -1)

        msg = ': '.join((*args, str(exc)))

        if errno == 0:
            self.logger.info(msg, exc_info=self.options['debug'] and exc)
        elif not isinstance(exc, asyncio.CancelledError):
            self.logger.error(msg, exc_info=self.options['debug'] and exc)

    def send_continue(self):
        if self.request is None or not self.request.http_continue:
            return

        if self.request.content_length > self.options['client_max_body_size']:
            raise ExpectationFailed

        self.transport.write(
            b'HTTP/%s 100 Continue\r\n\r\n' % self.request.version
        )
        self.queue[1].put_nowait(None)

    async def _handle_request(self, request):
        response = request.create_response()
        timer = self.set_handler_timeout(self.options['app_handler_timeout'])

        try:
            if self.options['keepalive_timeout']:  # can be 0 on shutdown
                if b'connection' in request.headers:
                    if b'close' not in request.headers.getlist(b'connection'):
                        self.globals.connections.add(self)
                        request.http_keepalive = True
                elif request.version == b'1.1':
                    self.globals.connections.add(self)
                    request.http_keepalive = True

            if b'transfer-encoding' in request.headers:
                if (request.version != b'1.1' or
                        b'chunked' not in request.transfer_encoding):
                    raise BadRequest('unexpected Transfer-Encoding')

                request.has_body = True

            if b'content-length' in request.headers:
                if (request.has_body or
                        len(request.headers[b'content-length']) != 1):
                    raise BadRequest('ambiguous Content-Length')

                try:
                    request.content_length = parse_int(
                        request.headers[b'content-length'][0]
                    )
                except ValueError as exc:
                    raise BadRequest('bad Content-Length') from exc

                request.has_body = request.content_length != 0

            if (request.has_body and b'expect' in request.headers and
                    request.headers[b'expect'][0].lower() == b'100-continue'):
                # we can handle continue later after the route is found
                # by checking this state
                request.http_continue = True

            # successfully got header,
            # clear either the request or keepalive timeout
            if not self.events['request'].done():
                self.events['request'].set_result(request)

            await self.request_received(request, response)
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as exc:
            if request.protocol is not None:
                data = None

                try:
                    data = await self.error_received(exc, response)
                except Exception as e:
                    exc = e
                finally:
                    await response.handle_exception(exc, data=data)
        finally:
            timer.cancel()
            await request.handler_exit()

    async def _receive_data(self):
        if 'request' not in self.events:
            return

        request = await self.events['request']
        excess = 0

        if not request.upgraded:
            request.body_size += len(self._recv_buf)

            if request.body_size > request.content_length > -1:
                excess = request.body_size - request.content_length

        while len(self._recv_buf) > excess:
            data = self._recv_buf[:min(self.options['buffer_size'],
                                       len(self._recv_buf) - excess)]

            if await self.put_to_queue(data, rate=self.options['upload_rate']):
                del self._recv_buf[:len(data)]
            else:
                del self._recv_buf[:]
                return

        # maybe resume reading, or close
        if 'receive' in self.events:
            del self.events['receive']

            if request.upgraded:
                if self in self.globals.connections:
                    self.transport.resume_reading()
            elif request.body_size > self.options['client_max_body_size']:
                self.close(BadRequest('payload too large'))
            elif request.body_size >= request.content_length > -1:
                self.queue[0].put_nowait(None)
            elif request.has_body and not request.eof():
                self.transport.resume_reading()

    def data_received(self, data):
        if not data:
            return

        self._recv_buf.extend(data)

        if self.request is None:
            header_size = self._recv_buf.find(b'\r\n\r\n') + 2

            if header_size > self.options['client_max_header_size']:
                self.close(BadRequest('request header too large'))
            elif header_size != 1:
                header = HTTPHeader().parse(self._recv_buf, header_size)
                del self._recv_buf[:header_size + 2]

                if header.is_request:
                    self.request = HTTPRequest(self, header)
                    task = self.app.create_task(
                        self._handle_request(self.request)
                    )

                    self.handlers.add(task)
                    task.add_done_callback(self.handlers.discard)
                else:
                    self.close(BadRequest('bad request: not a request'))
            elif len(self._recv_buf) > self.options['client_max_header_size']:
                self.close(BadRequest('bad request'))

            if self.request is None or not self._recv_buf:
                return

        # resumed in _receive_data or _send_data
        self.transport.pause_reading()

        if 'receive' not in self.events:
            self.events['receive'] = self.create_task(self._receive_data())

    def resume_writing(self):
        if 'send' in self.events and not self.events['send'].done():
            self.events['send'].set_result(None)

    def set_watermarks(self, high=65536, low=8192):
        if self.transport is not None:
            self.transport.set_write_buffer_limits(high=high, low=low)

    async def _send_data(self):
        try:
            while self.queue:
                data = await self.queue[1].get()

                if data is None:
                    # close the transport, unless keepalive is enabled
                    if self.request is not None:
                        if self.request.http_continue:
                            self.request.http_continue = False
                            continue

                        if self.request.http_keepalive:
                            self.request.http_keepalive = False

                            if self in self.globals.connections:
                                if not self.request.upgraded:
                                    await self._handle_keepalive()

                                self.transport.resume_reading()
                                continue

                            self.logger.info(
                                'keepalive connection dropped: %d', self.fileno
                            )

                        del self._recv_buf[:]
                        self.request.clear()

                    self.close()
                    return

                # send data
                write_buffer_size = self.transport.get_write_buffer_size()
                low, high = self.transport.get_write_buffer_limits()

                if write_buffer_size > high:
                    self.logger.debug(
                        '%d exceeds the current watermark limits '
                        '(high=%d, low=%d)',
                        write_buffer_size, high, low
                    )
                    self.events['send'] = self.loop.create_future()

                    await self.set_timeout(
                        self.events['send'],
                        timeout=self.options['request_timeout'],
                        timeout_cb=self.send_timeout
                    )

                    if self.transport is None:
                        return

                self.transport.write(data)
        except asyncio.CancelledError:
            self.close()
        except Exception as exc:
            self.close(exc)

    async def _handle_keepalive(self):
        if self.request.has_body and not self.request.eof():
            self.logger.info('request body was not fully consumed')
            self.request.clear()
            self.close()
            return

        if 'receive' in self.events:
            # waits for all incoming data to enter the queue
            await self.events['receive']

        self.events['request'] = self.loop.create_future()

        self.app.add_task(self.create_task(
            self.set_timeout(self.events['request'],
                             timeout=self.options['keepalive_timeout'],
                             timeout_cb=self.keepalive_timeout)
        ))

        # reset. so the next data in data_received will be considered as
        # a fresh http request (not a continuation data)
        self.request.clear()
        self.request = None
        self.server.clear()
        self.server.update(self.__dict__)
        self.queue[1].queue.clear()

        if self._recv_buf:
            self.queue[0].put_nowait(self._recv_buf[:])
            del self._recv_buf[:]

        while self.queue[0].qsize():
            # this data is supposed to be the next header
            self.data_received(self.queue[0].get_nowait())

    def connection_lost(self, _):
        while self.context.tasks:
            task = self.context.tasks.pop()

            try:
                getattr(task, 'cancel', task)()
            except Exception as exc:
                self.print_exception(exc, 'connection_lost')

        while self.queue:
            self.queue.pop().queue.clear()

        self.request = None

        self.context.clear()
        self.server.clear()
        self.events.clear()

        self.set_handler_timeout(self.options['app_close_timeout'])
        self.globals.connections.discard(self)



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

from tremolo.utils import parse_fields


class TremoloException(Exception):
    message = 'TremoloException'

    def __init__(self, *args):
        self.args = args

    def __str__(self):
        if self.args:
            errno, *msg = self.args

            if msg and isinstance(errno, int):
                return '%d: %s' % (errno, ' '.join(map(str, msg)))

            return ' '.join(map(str, self.args))

        return self.message


class HTTPException(TremoloException):
    code = 500
    message = 'Internal Server Error'
    content_type = 'text/html; charset=utf-8'

    def __new__(cls, *args, cause=None, **kwargs):
        if isinstance(cause, HTTPException):
            return cause

        if cause is not None and cls is HTTPException:
            if isinstance(cause, TimeoutError):
                cls = RequestTimeout
            else:
                cls = InternalServerError

        return super().__new__(cls)

    def __init__(self, *args, code=None, message=None, content_type=None,
                 cause=None, **options):
        if isinstance(code, int):
            self.code = code

        if isinstance(message, str):
            self.message = message

        if isinstance(content_type, str):
            self.content_type = content_type

        if isinstance(cause, Exception):
            if cause is self:
                if not options:
                    options = self.options
            else:
                self.__cause__ = cause

            if cause.args and not args:
                args = cause.args

        self.args = args
        self.options = options

    @property
    def encoding(self):
        for key, value in parse_fields(self.content_type.encode('latin-1')):
            if key == b'charset' and value:
                return value.decode('latin-1')

        return 'utf-8'


class HTTPRedirect(HTTPException):
    code = 302
    message = 'Found'


class BadRequest(HTTPException):
    code = 400
    message = 'Bad Request'


class Unauthorized(HTTPException):
    code = 401
    message = 'Unauthorized'


class Forbidden(HTTPException):
    code = 403
    message = 'Forbidden'


class NotFound(HTTPException):
    code = 404
    message = 'Not Found'


class MethodNotAllowed(HTTPException):
    code = 405
    message = 'Method Not Allowed'


class RequestTimeout(HTTPException):
    code = 408
    message = 'Request Timeout'


class PreconditionFailed(HTTPException):
    code = 412
    message = 'Precondition Failed'


class PayloadTooLarge(HTTPException):
    code = 413
    message = 'Payload Too Large'


class RangeNotSatisfiable(HTTPException):
    code = 416
    message = 'Range Not Satisfiable'


class ExpectationFailed(HTTPException):
    code = 417
    message = 'Expectation Failed'


class TooManyRequests(HTTPException):
    code = 429
    message = 'Too Many Requests'


class InternalServerError(HTTPException):
    pass


class ServiceUnavailable(HTTPException):
    code = 503
    message = 'Service Unavailable'


class WebSocketException(HTTPException):
    code = 1011

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)


class WebSocketClientClosed(WebSocketException):
    code = 1005


class WebSocketServerClosed(WebSocketException):
    pass



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto


class KeepAliveConnections(dict):
    def __init__(self, maxlen=512):
        if not isinstance(maxlen, int) or maxlen < 1:
            raise ValueError(
                'maxlen must be an integer greater than or equal to 1, '
                'got %s' % repr(maxlen)
            )

        self._maxlen = maxlen

    def __repr__(self):
        return self.keys().__repr__()

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)

        if self.__len__() > self._maxlen:
            del self[self.__iter__().__next__()]

    def add(self, item):
        self[item] = None

    def discard(self, item):
        self.pop(item, None)



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

from .exceptions import BadRequest, MethodNotAllowed, NotFound
from .lib.http_protocol import HTTPProtocol
from .lib.http_response import CONNECTIONS
from .lib.sse import SSE
from .lib.websocket import WebSocket


class HTTPServer(HTTPProtocol):
    async def _connection_made(self):
        for _, func in self.app.hooks['connect']:
            if await func(**self.server):
                break

    async def _connection_lost(self, exc):
        try:
            for _, func in reversed(self.app.hooks['close']):
                if await func(**self.server):
                    break
        except Exception as e:
            self.print_exception(e, '_connection_lost')
        finally:
            super().connection_lost(exc)

    def connection_made(self, transport):
        super().connection_made(transport)

        if self.app.hooks['connect']:
            self.events['connect'] = self.create_task(self._connection_made())
            self.app.add_task(self.events['connect'])

    def connection_lost(self, exc):
        if self.app.hooks['close']:
            task = self.app.create_task(self._connection_lost(exc))
            self.loop.call_at(
                self.loop.time() + self.options['app_close_timeout'],
                task.cancel
            )
        else:
            super().connection_lost(exc)

    async def run_middlewares(self, name, reverse=False, step=1):
        if self.is_closing():
            return

        middlewares = self.app.middlewares[self.server['prefix']][name]

        if reverse and self.server['next'] != -1:
            self.server['next'] = len(middlewares) - 1
            step = -1

        while -1 < self.server['next'] < len(middlewares):
            middleware = middlewares[self.server['next']]

            if await self._handle_middleware(middleware[1], middleware[2]):
                if reverse:
                    self.server['next'] = -1
                else:
                    self.server['next'] = len(middlewares)

                return True

            self.server['next'] += step

    async def _handle_middleware(self, func, kwargs):
        response = self.server['response']
        options = response.request.context.options
        options.update(kwargs)

        data = await func(**self.server)

        if data is None:
            return False

        if isinstance(data, (bytes, bytearray, str, tuple)):
            if 'status' in options:
                response.set_status(*options['status'])

            if 'content_type' in options:
                response.set_content_type(options['content_type'])

            encoding = ('utf-8',)

            if isinstance(data, tuple):
                data, *encoding = data + encoding

            if isinstance(data, str):
                data = data.encode(encoding[0])

            await response.end(data)
        else:
            self.logger.debug('middleware %s has exited with the connection '
                              'possibly left open', func.__name__)

        return True

    async def _handle_response(self, func, kwargs):
        response = self.server['response']
        request = response.request
        options = request.context.options
        options.update(kwargs)

        options.setdefault('rate', self.options['download_rate'])
        options.setdefault('buffer_size', self.options['buffer_size'])

        if not request.has_body:
            if ('websocket' in options and self.options['ws'] and
                    b'sec-websocket-key' in request.headers and
                    b'upgrade' in request.headers and
                    request.headers[b'upgrade'][0].lower() == b'websocket'):
                self.server['websocket'] = WebSocket(request, response)

            if 'sse' in options:
                self.server['sse'] = SSE(request, response)

        if 'status' in options:
            response.set_status(*options['status'])

        if 'content_type' in options:
            response.set_content_type(options['content_type'])

        try:
            coro = func(func=func, kwargs=kwargs, **self.server)
        except TypeError:  # doesn't accept extra **kwargs
            coro = func(**{k: self.server.get(k, kwargs[k]) for k in kwargs})
        finally:
            self.server.pop('self', None)  # avoid double self in middleware

        next_data = getattr(coro, '__anext__', None)

        if next_data:
            data = await next_data()
        else:
            data = await coro

            if data is None:
                response.close(keepalive=response.headers_sent())
                return

            if not isinstance(data, (bytes, bytearray, str, tuple)):
                self.logger.debug('handler %s has exited with the connection '
                                  'possibly left open', func.__name__)
                return

        status = response.get_status()
        no_content = status[0] in (204, 205, 304) or 100 <= status[0] < 200
        response.http_chunked = options.get(
            'chunked', request.version == b'1.1' and not no_content
        )
        response.set_status(*status)

        if not no_content:
            response.set_header(b'Content-Type', response.content_type)

        if response.http_chunked:
            response.set_header(b'Transfer-Encoding', b'chunked')

        if next_data:
            response.set_connection(status[0])

            if await self.run_middlewares('response', reverse=True):
                return

            if request.method == b'HEAD' or no_content:
                await response.write()
                return

            buffer_min_size = options['buffer_size'] // 2
            self.set_watermarks(high=options['buffer_size'] * 4,
                                low=buffer_min_size)

            if options.get('stream', True):
                buffer_min_size = None

            if data != b'':
                await response.write(data,
                                     rate=options['rate'],
                                     buffer_size=options['buffer_size'],
                                     buffer_min_size=buffer_min_size)

            while True:
                try:
                    data = await next_data()

                    if data == b'':
                        continue

                    await response.write(data,
                                         rate=options['rate'],
                                         buffer_size=options['buffer_size'],
                                         buffer_min_size=buffer_min_size)
                except StopAsyncIteration:
                    await response.write(b'', buffer_min_size=buffer_min_size)
                    break
        else:
            encoding = ('utf-8',)

            if isinstance(data, tuple):
                data, *encoding = data + encoding

            if isinstance(data, str):
                data = data.encode(encoding[0])

            if not (no_content or response.http_chunked):
                response.set_header(b'Content-Length', b'%d' % len(data))

            response.set_header(
                b'Connection', CONNECTIONS[request.http_keepalive]
            )

            if await self.run_middlewares('response', reverse=True):
                return

            if request.method == b'HEAD' or no_content:
                await response.write()
                return

            if data != b'':
                self.set_watermarks(high=options['buffer_size'] * 4,
                                    low=options['buffer_size'] // 2)
                await response.write(data,
                                     rate=options['rate'],
                                     buffer_size=options['buffer_size'])

            await response.write(b'')

        response.close(keepalive=True)

    async def request_received(self, request, response):
        self.server['request'] = request
        self.server['response'] = response
        self.server['next'] = 0

        if self.app.hooks['connect']:
            await self.events['connect']

        path = request.path.strip(b'/')

        if path == b'':
            key = 1
            self.server['prefix'] = ()
        else:
            parts = path.split(b'/', 254)
            length = len(parts)
            key = bytes([length]) + parts[0]

            while length >= 0:
                self.server['prefix'] = tuple(parts[:length])

                if self.server['prefix'] in self.app.middlewares:
                    break

                length -= 1

        if await self.run_middlewares('request'):
            await self.run_middlewares('response', reverse=True)
            return

        if not request.is_valid:
            raise BadRequest

        if key not in self.app.routes:
            key = parts[0]

        if request.method == b'HEAD':
            method = b'GET'
        else:
            method = request.method

        methods = set()

        while True:
            if key in self.app.routes:
                i = len(self.app.routes[key])

                while i > 0:
                    i -= 1
                    p, func, kwargs, options = self.app.routes[key][i]
                    m = p.search(request.url)

                    if m:
                        if (key == -1 and path != b'' and
                                p.pattern.startswith(b'^/' + parts[0])):
                            if parts[0] in self.app.routes:
                                self.app.routes[parts[0]].append(
                                    self.app.routes[-1].pop(i)
                                )
                            else:
                                self.app.routes[parts[0]] = [
                                    self.app.routes[-1].pop(i)
                                ]

                        matches = m.groupdict()
                        request.params['path'] = matches or m.groups()

                        if 'self' in kwargs:
                            methods.add(func.__name__.upper().encode())

                            if method not in methods:
                                continue

                            if callable(kwargs['self']):
                                matches['self'] = kwargs['self'](**options)

                        for k in matches:
                            self.server.setdefault(k, matches[k])

                        await self._handle_response(func, kwargs)
                        return

            if key == -1 or method in methods:
                break

            key = -1

        if methods:
            raise MethodNotAllowed(allow=b', '.join(methods))

        raise NotFound



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

from traceback import TracebackException

from .utils import html_escape


async def index(status=(503, b'Service Unavailable')):
    return b'Service Unavailable'


async def error_400(exc, **_):
    return str(exc)


async def error_404(request, **_):
    return (
        b'<!DOCTYPE html><html lang="en"><head><meta name="viewport" '
        b'content="width=device-width, initial-scale=1.0" />'
        b'<title>404 Not Found</title>'
        b'<style>body { max-width: 600px; margin: 0 auto; padding: 1%%; '
        b'font-family: sans-serif; line-height: 1.5em; }</style></head>'
        b'<body><h1>Not Found</h1><p>Unable to find handler for %s.</p><hr />'
        b'<address title="Powered by Tremolo">%s</address></body></html>' %
        (html_escape(request.path), request.server.globals.info['server_name'])
    )


async def error_405(exc, **_):
    return str(exc)


async def error_500(request, exc, **_):
    if request.protocol is None:
        return

    if request.protocol.options['debug']:
        te = TracebackException.from_exception(exc)
        return '<ul><li>%s</li></ul>' % '</li><li>'.join(
            html_escape(line) for line in te.format()
        )

    return str(exc)



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

from .http_exceptions import InternalServerError


class SSE:
    def __init__(self, request, response):
        self.request = request
        self.response = response

        response.set_content_type(b'text/event-stream')
        response.set_header(b'Cache-Control', b'no-cache, must-revalidate')
        response.set_header(b'Expires', b'Thu, 01 Jan 1970 00:00:00 GMT')

    async def send(self, data, event=None, event_id=None, retry=0):
        if isinstance(data, str):
            data = data.strip('\n').encode('utf-8')
        else:
            data = data.strip(b'\n')

        if b'\n' in data:
            data = data.replace(b'\n', b'\ndata: ')

        for name, value in ((b'event', event), (b'id', event_id)):
            if value:
                if isinstance(value, str):
                    value = value.encode('utf-8')

                if b'\n' in value:
                    raise InternalServerError

                data += b'\n%s: %s' % (name, value)

        if retry:
            data += b'\nretry: %d' % retry

        await self.response.write(b'data: %s\n\n' % data)

    async def close(self):
        try:
            await self.response.write(b'')
            self.response.close(keepalive=True)
        except RuntimeError:
            pass



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import asyncio

from collections import deque


class Queue:
    def __init__(self, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self.queue = deque()
        self._getters = deque()

    def qsize(self):
        return len(self.queue)

    def put_nowait(self, item):
        while self._getters:
            fut = self._getters.popleft()

            if not fut.done():
                fut.set_result(item)
                return

        self.queue.append(item)

    async def get(self, timeout=None):
        if self._getters:
            await self._getters[-1]

        try:
            return self.queue.popleft()
        except IndexError:
            fut = self._loop.create_future()
            self._getters.append(fut)

            if timeout is not None:
                timer = self._loop.call_at(self._loop.time() + timeout,
                                           fut.cancel)

            try:
                return await fut
            except asyncio.CancelledError as exc:
                try:
                    self._getters.remove(fut)
                except ValueError:
                    pass

                raise exc from None
            finally:
                if timeout is not None:
                    timer.cancel()

    def get_nowait(self):
        return self.queue.popleft()



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import asyncio

from http import HTTPStatus
from urllib.parse import unquote_to_bytes

from .exceptions import (
    ASGIConnectionError,
    BadRequest,
    Forbidden,
    WebSocketClientClosed,
    WebSocketServerClosed
)
from .lib.http_protocol import HTTPProtocol
from .lib.websocket import WebSocket


class ASGIServer(HTTPProtocol):
    def connection_made(self, transport):
        super().connection_made(transport)

        self.events['close'] = self.loop.create_future()

    def connection_lost(self, exc):
        if not self.events['close'].done():
            self.events['close'].set_result(exc)

        super().connection_lost(exc)

    async def request_received(self, request, response):
        if not request.is_valid:
            raise BadRequest

        scope = {
            'asgi': {'version': '3.0', 'spec_version': '2.4'},
            'root_path': self.options['root_path'],
            'server': self.globals.info['server'],
            'client': request.client,
            'http_version': request.version.decode('latin-1'),
            'path': unquote_to_bytes(request.path).decode('latin-1'),
            'raw_path': request.path,
            'query_string': request.query_string,
            'headers': request.header.getheaders(),
            'state': self.options['state'].copy()
        }

        # provide direct access to server objects
        self.server['request'] = request
        self.server['response'] = response
        scope['state']['server'] = self.server

        if (self.options['ws'] and b'sec-websocket-key' in request.headers and
                b'upgrade' in request.headers and
                request.headers[b'upgrade'][0].lower() == b'websocket'):
            scope['type'] = 'websocket'
            scope['scheme'] = b'wss' if request.scheme == b'https' else b'ws'
            scope['subprotocols'] = [
                value.decode('latin-1') for value in
                request.headers.getlist(b'sec-websocket-protocol')]
        else:
            scope['type'] = 'http'
            scope['method'] = request.method.decode('latin-1')
            scope['scheme'] = request.scheme.decode('latin-1')

        app = ASGIAppWrapper(self, self.options['app'], scope, response)

        try:
            await app

            if 'close' in self.events and not self.events['close'].cancelled():
                if not self.events['close'].done():
                    self.logger.info('handler exited early (no close?)')

                raise await self.events['close']
        except (asyncio.CancelledError, Exception) as exc:
            if scope['type'] == 'websocket' and request.upgraded:
                exc = WebSocketServerClosed(cause=exc)

            await response.handle_exception(exc, 'app')
        finally:
            scope.clear()


class ASGIAppWrapper:
    def __init__(self, protocol, app, scope, response):
        self.protocol = protocol
        self.app = app
        self.scope = scope
        self.loop = protocol.loop
        self.logger = protocol.logger
        self.request = response.request
        self.response = response
        self.websocket = None

    def __await__(self):
        return self.app(self.scope, self.receive, self.send).__await__()

    async def receive(self):
        if self.scope['type'] == 'websocket':
            # initially, the Request.upgraded value is False
            # it will become True later
            # after the response status is set to 101:
            # Response.set_status(101) in WebSocket.accept()
            if not self.request.upgraded and self.websocket is None:
                self.websocket = WebSocket(self.request, self.response)

                return {'type': 'websocket.connect'}

            try:
                payload = await self.websocket.receive()

                if isinstance(payload, str):
                    return {
                        'type': 'websocket.receive',
                        'text': payload
                    }

                return {
                    'type': 'websocket.receive',
                    'bytes': payload
                }
            except Exception as exc:
                code = 1005

                if self.websocket is None or self.protocol.is_closing():
                    self.logger.info(
                        'calling receive() after the connection is closed'
                    )
                else:
                    if isinstance(exc, WebSocketClientClosed):
                        code = exc.code
                    else:
                        self.protocol.print_exception(exc, 'receive')

                    if isinstance(exc, WebSocketServerClosed):
                        await self.websocket.close(exc.code)
                    else:
                        await self.websocket.close(1011 if code == 1005
                                                   else 1000)

                    self.websocket = None
                    self.response = None

                    self.protocol.request = None  # force handler timeout
                    self.protocol.set_handler_timeout(
                        self.protocol.options['app_close_timeout']
                    )

                return {
                    'type': 'websocket.disconnect',
                    'code': code
                }

        try:
            data = await self.request.read()

            if not data:
                self.request = None

            return {
                'type': 'http.request',
                'body': data,
                'more_body': bool(data)
            }
        except Exception as exc:
            if self.response is None or self.protocol.is_closing():
                self.logger.info(
                    'calling receive() after the connection is closed'
                )
            else:
                if self.request is None:
                    # delay http.disconnect. see:
                    # https://github.com/nggit/tremolo/issues/202
                    if self.protocol.events['close'].done():
                        self.protocol.events[
                            'close'] = self.loop.create_future()

                    self.logger.info(
                        'calling receive() when there is no more body'
                    )
                    try:
                        await self.protocol.events['close']
                    except asyncio.CancelledError:
                        if self.response is not None:  # not a wake up
                            raise
                else:
                    await self.response.handle_exception(exc, 'receive')
                    self.response = None

                    # break the while loop (if any)
                    raise asyncio.CancelledError from None

                self.protocol.set_handler_timeout(
                    self.protocol.options['app_close_timeout']
                )

            return {'type': 'http.disconnect'}

    async def send(self, data):
        try:
            if data['type'] in ('http.response.start', 'websocket.accept'):
                if self.response.line is not None:
                    raise RuntimeError('already started or accepted')

                # websocket doesn't have this
                if 'status' in data:
                    self.response.set_status(data['status'],
                                             HTTPStatus(data['status']).phrase)

                if 'headers' in data:
                    for header in data['headers']:
                        name = header[0].lower()

                        if name == b'content-type':
                            self.response.set_content_type(header[1])
                            continue

                        if name in (b'date', b'server', b'transfer-encoding'):
                            # disallow apps from changing them,
                            # as they are managed by Tremolo
                            continue

                        if name == b'connection':
                            if header[1].lower() == b'close':
                                self.response.request.http_keepalive = False

                            continue

                        if name == b'content-length':
                            # will disable http chunked in the
                            # self.response.write()
                            self.response.http_chunked = False

                        self.response.append_header(*header)

                # websocket has this
                if 'subprotocol' in data and data['subprotocol']:
                    self.response.set_header(
                        b'Sec-WebSocket-Protocol', data['subprotocol']
                    )
            elif self.response.line is None:
                if data['type'] == 'websocket.close':
                    self.websocket = None
                    raise Forbidden('connection rejected')

                raise RuntimeError('has not been started or accepted')

            if data['type'] == 'http.response.body' and self.websocket is None:
                if 'body' in data and data['body'] != b'':
                    await self.response.write(
                        data['body'],
                        rate=self.protocol.options['download_rate'],
                        buffer_size=self.protocol.options['buffer_size']
                    )

                if 'more_body' not in data or not data['more_body']:
                    await self.response.write(b'')
                    self.response.close(keepalive=True)
                    self.request = None  # disallows further receive()
                    self.response = None  # disallows further send()
                    self.protocol.events['close'].cancel()  # wake up receive()
            elif data['type'] == 'websocket.send' and self.websocket:
                if 'bytes' in data and data['bytes']:
                    await self.websocket.send(data['bytes'])
                elif 'text' in data and data['text']:
                    await self.websocket.send(data['text'], opcode=1)
            elif data['type'] == 'websocket.accept' and self.websocket:
                await self.websocket.accept()
            elif data['type'] == 'websocket.close' and self.websocket:
                await self.websocket.close(data.get('code', 1000))
                self.websocket = None
                self.response = None
            elif data['type'] != 'http.response.start':
                raise RuntimeError('unexpected ASGI message type')
        except Exception as exc:
            if self.response is None or self.protocol.is_closing():
                raise ASGIConnectionError(
                    'calling send() after the connection is closed'
                ) from None

            self.protocol.events['close'].set_result(exc)  # backup
            self.response = None
            raise  # this may be swallowed by the app, so we have a backup



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import sys

from tremolo import __version__, Tremolo
from tremolo.utils import parse_args

server = Tremolo()


def usage(**context):
    print('Usage: python3 -m tremolo [OPTIONS] APP')
    print()
    print('Example:')
    print('  python3 -m tremolo example:app')
    print('  python3 -m tremolo /path/to/example.py')
    print('  python3 -m tremolo /path/to/example.py:myapp')
    print('  python3 -m tremolo --debug --port 8080 example:app')
    print()
    print('Options:')
    print('  --host                    Listen host. Defaults to "127.0.0.1"')
    print('  --port                    Listen port. Defaults to 8000')
    print('  --bind                    Address to bind.')
    print('                            Instead of using --host or --port')
    print('                            E.g. "127.0.0.1:8000" or "/tmp/file.sock"')  # noqa: E501
    print('                            Multiple binds can be separated by commas')  # noqa: E501
    print('                            E.g. "127.0.0.1:8000,:8001"')
    print('  --worker-num              Number of worker processes. Defaults to 1')  # noqa: E501
    print('  --thread-pool-size        Number of executor threads per process')  # noqa: E501
    print('                            Defaults to 5')
    print('  --limit-memory            Restart the worker if this limit (in KiB) is reached')  # noqa: E501
    print('                            (Linux-only). Defaults to 0 or unlimited')  # noqa: E501
    print('  --backlog                 Maximum number of pending connections')
    print('                            Defaults to 100')
    print('  --ssl-cert                SSL certificate location')
    print('                            E.g. "/path/to/fullchain.pem"')
    print('  --ssl-key                 SSL private key location')
    print('                            E.g. "/path/to/privkey.pem"')
    print('  --debug                   Enable debug mode')
    print('                            Intended for development')
    print('  --experimental            Enable experimental features')
    print('  --reload                  Enable auto reload on code changes')
    print('                            Intended for development')
    print('  --no-ws                   Disable built-in WebSocket support')
    print('  --ws-max-payload-size     Maximum payload size for the built-in WebSocket')  # noqa: E501
    print('                            Defaults to 2 * 1048576, or 2MiB')
    print('  --log-level               Defaults to "DEBUG". See')
    print('                            https://docs.python.org/3/library/logging.html#levels')  # noqa: E501
    print('  --log-fmt                 Python\'s log format. If empty defaults to "%(message)s"')  # noqa: E501
    print('  --loop                    A fully qualified event loop name')
    print('                            E.g. "asyncio" or "asyncio.SelectorEventLoop"')  # noqa: E501
    print('                            It expects the respective module to already be present')  # noqa: E501
    print('  --download-rate           Limits the sending speed to the client')
    print('                            Defaults to 1048576, which means 1MiB/s')  # noqa: E501
    print('  --upload-rate             Limits the upload / POST speed')
    print('                            Defaults to 1048576, which means 1MiB/s')  # noqa: E501
    print('  --buffer-size             Defaults to 16384, or 16KiB')
    print('  --client-max-body-size    Defaults to 2 * 1048576, or 2MiB')
    print('  --client-max-header-size  Defaults to 8192, or 8KiB')
    print('  --max-queue-size          Maximum number of buffers in the queue')
    print('                            Defaults to 128')
    print('  --request-timeout         Defaults to 30 (seconds)')
    print('  --keepalive-timeout       Defaults to 30 (seconds)')
    print('  --keepalive-connections   Maximum number of keep-alive connections')  # noqa: E501
    print('                            Defaults to 512 (connections/worker)')
    print('  --app-handler-timeout     Kill the app if it takes too long to finish')  # noqa: E501
    print('                            Upgraded connection/scope will not be affected')  # noqa: E501
    print('                            Defaults to 120 (seconds)')
    print('  --app-close-timeout       Kill the app if it does not exit within this timeframe,')  # noqa: E501
    print('                            from when the client is disconnected')
    print('                            Defaults to 30 (seconds)')
    print('  --shutdown-timeout        Maximum number of seconds to wait after SIGTERM is')  # noqa: E501
    print('                            sent to a worker process. Defaults to 30 (seconds)')  # noqa: E501
    print('  --server-name             Set the "Server" field in the response header')  # noqa: E501
    print('  --root-path               Set the ASGI root_path. Defaults to ""')
    print('  --version                 Print the tremolo version and exit')
    print('  --help                    Show this help and exit')
    return 0


def bind(value='', **context):
    context['options']['host'] = None

    try:
        for bind in value.split(','):
            if ':\\' not in bind and ':' in bind:
                host, port = bind.rsplit(':', 1)
                server.listen(int(port), host=host.strip('[]') or None)
            else:
                server.listen(bind)
    except ValueError:
        print('Invalid --bind value "%s"' % value)
        return 1


def version(**context):
    print(
        'tremolo %s (%s %d.%d.%d, %s)' %
        (__version__,
         sys.implementation.name,
         *sys.version_info[:3],
         sys.platform)
    )
    return 0


if __name__ == '__main__':
    options = parse_args(help=usage, bind=bind, version=version)

    if sys.argv[-1] != sys.argv[0] and not sys.argv[-1].startswith('-'):
        options['app'] = sys.argv[-1]

    if 'app' not in options:
        print('You must specify APP. Use "--help" for help')
        sys.exit(1)

    server.run(**options)



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

from tremolo.utils import parse_fields


def is_safe(data):
    return not (b'\r' in data or b'\n' in data or 0 in data)


def iter_lines(data, limit=65536):
    start = 0

    while start < limit:
        end = data.find(b'\r\n', start, limit)

        if end == -1:
            if start < len(data):
                yield bytes(data[start:limit])

            break

        yield bytes(data[start:end])
        start = end + 2


class Headers(dict):
    def getlist(self, name, separator=b',', split=None):
        result = []

        if name in self:
            for value in self[name]:
                for v in parse_fields(value, separator, split):
                    result.append(v)

        return result


class HTTPHeader:
    __slots__ = ('is_valid', 'line', 'method', 'url', 'version', 'headers')

    def __init__(self):
        self.is_valid = False
        self.line = None
        self.method = None
        self.url = None
        self.version = None
        self.headers = Headers()

    @property
    def is_request(self):
        return not (self.method is None or self.url is None)

    def parse(self, data, header_size=0, max_lines=100, max_line_size=8190):
        if self.line is not None:
            return self.__class__().parse(data, header_size,
                                          max_lines, max_line_size)

        if not data:
            return self

        lines = iter_lines(data, header_size or len(data))
        self.line = lines.__next__()

        if len(self.line) > max_line_size or not is_safe(self.line):
            return self

        path_end = self.line.rfind(b' HTTP/')

        if path_end > 0:
            try:
                self.method, self.url = self.line[:path_end].split(b' ', 1)
                self.version = self.line[path_end + 6:]
            except ValueError:
                self.method = b''
                self.url = b''

            if self.version in (b'1.1', b'1.0'):
                self.is_valid = True
            else:
                self.version = b'1.0'

        while max_lines > 1:
            try:
                line = lines.__next__()
            except StopIteration:
                break

            if line == b'':
                break

            if len(line) > max_line_size or not is_safe(line):
                self.is_valid = False
                return self

            colon_pos = line.find(b':', 1)

            if colon_pos <= 0 or line[colon_pos - 1] == 32:
                self.is_valid = False
                break

            name = line[:colon_pos].lower()
            value = line[colon_pos + 1:].strip(b' \t')

            if name in self.headers:
                self.headers[name].append(value)
            else:
                self.headers[name] = [value]

            max_lines -= 1
        else:
            self.is_valid = False

        if self.is_request:
            if b'host' in self.headers:
                if len(self.headers[b'host']) != 1:
                    self.is_valid = False
            else:
                self.headers[b'host'] = [b'']

                if self.is_valid and self.version == b'1.1':
                    self.is_valid = False

        return self

    def getheaders(self):
        return [(k, v) for k in self.headers for v in self.headers[k]]

    def gethost(self):
        values = self.headers.get(b'x-forwarded-host',
                                  self.headers.get(b'host'))
        if values:
            return values[0]



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

from .tremolo import __version__, Tremolo, Tremolo as Application  # noqa: F401
from . import exceptions  # noqa: F401


def run(app, **options):
    if 'host' not in options:
        options['host'] = '127.0.0.1'

    if 'port' not in options:
        options['port'] = 8000

    Tremolo().run(app=app, **options)



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import os
import stat
import sys

from asyncio import run_coroutine_threadsafe as run_coroutine
from datetime import datetime, timezone
from html import escape
from inspect import iscoroutinefunction, isasyncgenfunction
from urllib.parse import unquote_to_bytes as unquote

__all__ = ['file_signature', 'getoptions', 'html_escape', 'is_async',
           'log_date', 'memory_usage', 'parse_args', 'parse_fields',
           'parse_int', 'print_logo', 'server_date', 'to_sync']


def file_signature(path):
    st = os.stat(path)

    return (stat.S_IFMT(st.st_mode), st.st_size, st.st_mtime)


def getoptions(func):
    options = {}
    argcount = func.__code__.co_argcount

    if func.__defaults__ is not None:
        argcount -= len(func.__defaults__)

    for name in func.__code__.co_varnames[:argcount]:
        options[name] = None

    for i, name in enumerate(func.__code__.co_varnames[
                                 argcount:func.__code__.co_argcount]):
        options[name] = func.__defaults__[i]

    return options


def html_escape(data):
    if isinstance(data, str):
        return escape(data)

    return (data.replace(b'&', b'&amp;')
            .replace(b'<', b'&lt;')
            .replace(b'>', b'&gt;')
            .replace(b'"', b'&quot;'))


def is_async(func):
    return iscoroutinefunction(func) or isasyncgenfunction(func)


def log_date():
    return datetime.now().strftime('[%Y-%m-%d %H:%M:%S]')


def memory_usage(pid=0):
    if not pid:
        pid = os.getpid()

    try:
        with open('/proc/%d/statm' % pid, 'r') as f:
            return int(f.read().split()[1]) * os.sysconf('SC_PAGESIZE') // 1024
    except FileNotFoundError:
        # non-Linux
        return -1


def parse_args(**callbacks):
    options = {'host': '127.0.0.1', 'port': 8000, 'ssl': {}}
    context = {'options': options}

    for i in range(len(sys.argv)):
        name = sys.argv[i - 1].lstrip('-').replace('-', '_')

        if sys.argv[i - 1] == '--no-ws':
            options['ws'] = False
        elif sys.argv[i - 1] in ('--debug', '--experimental', '--reload'):
            options[name] = True
        elif sys.argv[i - 1] in ('--host',
                                 '--log-level',
                                 '--log-fmt',
                                 '--loop',
                                 '--server-name',
                                 '--root-path'):
            options[name] = sys.argv[i]
        elif sys.argv[i - 1] in ('--port',
                                 '--worker-num',
                                 '--thread-pool-size',
                                 '--limit-memory',
                                 '--backlog',
                                 '--download-rate',
                                 '--upload-rate',
                                 '--buffer-size',
                                 '--ws-max-payload-size',
                                 '--client-max-body-size',
                                 '--client-max-header-size',
                                 '--max-queue-size',
                                 '--request-timeout',
                                 '--keepalive-timeout',
                                 '--keepalive-connections',
                                 '--app-handler-timeout',
                                 '--app-close-timeout',
                                 '--shutdown-timeout'):
            try:
                options[name] = int(sys.argv[i])
            except ValueError:
                print(
                    'Invalid %s value "%s". It must be a number' %
                    (sys.argv[i - 1], sys.argv[i])
                )
                sys.exit(1)
        elif sys.argv[i - 1] == '--ssl-cert':
            options['ssl']['cert'] = sys.argv[i]
        elif sys.argv[i - 1] == '--ssl-key':
            options['ssl']['key'] = sys.argv[i]
        elif sys.argv[i - 1].startswith('--'):
            if name in callbacks:
                code = callbacks[name](value=sys.argv[i], **context)

                if code is not None:
                    sys.exit(code)
            else:
                print('Unrecognized option "%s"' % sys.argv[i - 1])
                sys.exit(1)

    return options


def parse_fields(data, separator=b';', split=b'=', max_fields=100):
    if len(separator) != 1:
        raise ValueError('separator must be a single one-byte character')

    end = len(data)

    while max_fields > 0:
        start = data.rfind(separator, 0, end) + 1

        if split:
            name, _, value = data[start:end].partition(split)

            if name:
                yield (name.strip().lower(), unquote(value.strip(b' \t"')))
        else:
            yield data[start:end].strip().lower()

        if start == 0:
            break

        end = start - 1
        max_fields -= 1


def parse_int(string, base=10):
    if not isinstance(string, (bytes, bytearray)) or not 0 < len(string) <= 16:
        raise ValueError('string must be bytes-like and <=16 in length')

    if string == b'0':
        return 0

    if string.startswith(b'0'):
        raise ValueError('string cannot contain leading zeros')

    total = 0

    for c in string:
        total = total * base + int(chr(c), base)

    return total


def print_logo():
    BLUE = '\033[0;34m'
    RESET = '\033[0m'

    try:
        print(f'\n                                {BLUE}▃▃▇▇{RESET}')
    except UnicodeEncodeError:
        return

    print(f'  ▃                             {BLUE}▀▀  {RESET}')
    print(f'  █▃▃  ▃ ▃▃  ▃▃▃  ▃▃▃ ▃▃   ▃▃▃  {BLUE}▃▃▇▇{RESET}  ▃▃▃ ')
    print(f'  █    █▀   █▂▂▂█ █  █  █ █   █ {BLUE}▀▀  {RESET} █   █')
    print(f'  █    █    █▔▔▔▔ █  █  █ █   █ {BLUE}▃▃▇▇{RESET} █   █')
    print(f'   ▀▀▀ ▀     ▀▀▀  ▀  ▀  ▀  ▀▀▀  {BLUE}▀▀  {RESET}  ▀▀▀ \n')


def server_date():
    return datetime.now(timezone.utc).strftime(
        '%a, %d %b %Y %H:%M:%S GMT').encode('latin-1')


class AsyncToSyncWrapper:
    def __init__(self, obj, loop):
        self.__wrapped__ = obj
        self.loop = loop

    def __getattr__(self, name):
        attr = getattr(self.__wrapped__, name)

        if iscoroutinefunction(attr):
            def func(*args, **kwargs):
                return run_coroutine(attr(*args, **kwargs), self.loop).result()

            return func

        if isasyncgenfunction(attr):
            def generator(*args, **kwargs):
                agen = attr(*args, **kwargs)

                while True:
                    try:
                        fut = run_coroutine(agen.__anext__(), self.loop)
                        yield fut.result()
                    except StopAsyncIteration:
                        break

            return generator

        return attr


to_sync = AsyncToSyncWrapper



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import re

from functools import wraps

from . import handlers
from .utils import getoptions, is_async, to_sync


class Routes(dict):
    def __init__(self):
        self[0] = [
            (400, handlers.error_400, {}, {}),
            (401, None, {}, {}),
            (402, None, {}, {}),
            (403, None, {}, {}),
            (404, handlers.error_404, {}, {}),
            (405, handlers.error_405, {}, {})] + [
            (code, handlers.error_500 if code == 500 else None, {}, {})
            for code in range(406, 512)
        ]
        self[1] = [
            (b'^/+(?:\\?.*)?$', handlers.index, getoptions(handlers.index), {})
        ]
        self[-1] = []

    def add(self, func, path='/', kwargs=None, **options):
        if not kwargs:
            kwargs = getoptions(func)

        key = -1

        if path.endswith('$'):
            pattern = path.rstrip('$').encode('latin-1') + b'(?:\\?.*)?$'
        elif path.startswith('^'):
            pattern = path.encode('latin-1')
        else:
            path = path.split('?', 1)[0].strip('/').encode('latin-1')

            if path == b'':
                key = 1
                pattern = self[1][0][0]
            else:
                parts = path.split(b'/', 254)
                key = bytes([len(parts)]) + parts[0]
                pattern = b'^/+%s(?:/+)?(?:\\?.*)?$' % path

        if key != 1 and key in self:
            self[key].append((pattern, func, kwargs, options))
        else:
            self[key] = [(pattern, func, kwargs, options)]

    def compile(self, executor=None):
        for key in self:
            for i, (pattern, func, kwargs, options) in enumerate(self[key]):
                if func is None:
                    continue

                if isinstance(pattern, bytes):
                    pattern = re.compile(pattern)

                if executor is None or is_async(func):
                    wrapper = func
                else:
                    @wraps(func)
                    def wrapper(func, kwargs, request, response,
                                self=None, **server):
                        server['request'] = to_sync(request, server['loop'])
                        server['response'] = to_sync(response, server['loop'])

                        if self is None:
                            func = func.__wrapped__
                        else:  # use bound method
                            func = getattr(self, func.__name__)

                        return executor.submit(func, kwargs=server)

                self[key][i] = (pattern, wrapper, kwargs, options)



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto


class Context:
    def __init__(self, *args, **kwargs):
        self.__dict__.update(*args, **kwargs)

    def __repr__(self):
        return self.__dict__.__repr__()

    def __getattr__(self, name):  # clear(), get(), update(), etc.
        return getattr(self.__dict__, name)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __contains__(self, key):
        return self.__dict__.__contains__(key)


class WorkerContext(Context):
    _tasks = set()

    def __init__(self):
        self.__dict__['info'] = {}
        self.__dict__['options'] = {}

    @property
    def tasks(self):
        return self._tasks

    @property
    def info(self):
        return self.__dict__['info']

    @property
    def options(self):
        return self.__dict__['options']

    @property
    def connections(self):
        return self.__dict__.get('connections', None)

    @property
    def executor(self):
        return self.__dict__.get('executor', None)


class ConnectionContext(Context):
    __slots__ = ('_tasks',)  # won't appear in self.__dict__. safe from clear()

    def __init__(self):
        self._tasks = set()

    @property
    def tasks(self):
        return self._tasks

    @property
    def transport(self):
        return self.__dict__.get('transport', None)

    @property
    def client(self):
        if 'client' not in self.__dict__ and self.transport is not None:
            sock = self.transport.get_extra_info('socket')
            self.__dict__['client'] = sock.getpeername() or None

            if isinstance(self.__dict__['client'], tuple):
                self.__dict__['client'] = self.__dict__['client'][:2]

        return self.__dict__.get('client', None)


class RequestContext(Context):
    def __init__(self):
        self.__dict__['options'] = {}

    @property
    def options(self):
        return self.__dict__['options']



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import os
import time

from urllib.parse import parse_qs

from tremolo.utils import parse_fields, parse_int
from .http_exceptions import BadRequest, PayloadTooLarge
from .http_response import HTTPResponse
from .request import Request


class MultipartFile(dict):
    def __init__(self, files):
        self.files = files

    async def stream(self):
        if 'data' in self:
            yield self.pop('data')

            while not self['eof']:
                yield (await self.files.__anext__()).pop('data')


class HTTPRequest(Request):
    __slots__ = ('_ip', '_scheme', 'header', 'headers', 'is_valid',
                 'host', 'method', 'url', 'path', 'query_string', 'version',
                 'content_length', 'http_continue', 'http_keepalive',
                 'has_body', '_body', '_read_buf', '_stream', '_files')

    def __init__(self, protocol, header):
        super().__init__(protocol)

        self._ip = None
        self._scheme = None
        self.header = header
        self.headers = header.headers
        self.is_valid = header.is_valid
        self.host = header.gethost()
        self.method = header.method
        self.url = header.url
        self.path, _, self.query_string = self.url.partition(b'?')
        self.version = header.version

        self.content_length = -1
        self.http_continue = False
        self.http_keepalive = False
        self.has_body = False

        self._body = bytearray()
        self._read_buf = bytearray()
        self._stream = None
        self._files = None

    @property
    def ip(self):
        if not self._ip:
            values = self.headers.get(b'x-forwarded-for')
            ip = values[0].strip() if values else b''

            if ip == b'' and self.client is not None:
                self._ip = self.client[0].encode('utf-8')
            else:
                self._ip = ip.split(b',', 1)[0]

        return self._ip

    @property
    def scheme(self):
        if not self._scheme:
            values = self.headers.get(b'x-forwarded-proto')
            scheme = values[0].strip() if values else b''

            if scheme == b'' and self.is_secure:
                self._scheme = b'https'
            else:
                self._scheme = scheme or b'http'

        return self._scheme

    @property
    def content_type(self):
        # don't lower() content-type, as it may contain a boundary
        if b'content-type' in self.headers:
            return self.headers[b'content-type'][0]

        return b'application/octet-stream'

    @property
    def transfer_encoding(self):
        return self.headers.getlist(b'transfer-encoding')

    @property
    def upgraded(self):
        return self._body is None

    @upgraded.setter
    def upgraded(self, value):
        self._body = None
        del self._read_buf[:]

    def uid(self, length=32, *, ts_offset=0):
        if self.client is None:
            port = self.socket.fileno()
        else:
            port = self.client[1]  # 0 - 65535

        prefix = (
            int(time.time() + ts_offset).to_bytes(4, byteorder='big') +
            int.to_bytes((port << 16) | (os.getpid() & 0xffff),
                         4, byteorder='big')
        )[:length]  # 8 Bytes

        return prefix + os.urandom(length - len(prefix))

    def create_response(self):
        return HTTPResponse(self)

    def clear(self):
        self.headers.clear()

        if self._body:
            del self._body[:]

        del self._read_buf[:]

        super().clear()

    async def handler_exit(self):
        if self._files is not None:
            await self._files.aclose()

        if self._stream is not None:
            await self._stream.aclose()

    async def body(self, **kwargs):
        async for data in self.stream(**kwargs):
            self._body.extend(data)

        return self._body

    async def read(self, size=-1, *, timeout=None):
        return await self.recv(size, timeout=timeout, raw=False)

    def eof(self):
        return self.content_length == 0 and self._read_buf == b''

    async def recv(self, size=-1, *, timeout=None, raw=True):
        if size == 0 or self.eof():
            return b''

        if self._stream is None:
            self._stream = self.stream(timeout, raw)
        elif timeout is not None:  # update
            self.timeout = timeout

        if size == -1:
            return await self._stream.__anext__()

        try:
            while len(self._read_buf) < size:
                self._read_buf.extend(await self._stream.__anext__())

            return bytes(self._read_buf[:size])
        except StopAsyncIteration:
            return bytes(self._read_buf)
        finally:
            del self._read_buf[:size]

    async def stream(self, timeout=None, raw=False):
        if self.content_length == 0:
            return

        if self.http_continue:
            self.server.send_continue()

        if raw or self.content_length != -1:
            if (self.content_length >
                    self.server.options['client_max_body_size']):
                raise PayloadTooLarge

            async for data in super().recv(timeout):
                yield bytes(data)
        elif b'chunked' in self.transfer_encoding:
            buf = bytearray()
            agen = super().recv(timeout)
            paused = False
            bytes_unread = 0
            chunk_size = -1

            while True:
                if not paused:
                    buf.extend(await agen.__anext__())

                if bytes_unread > 2:
                    data = bytes(buf[:bytes_unread - 2])
                    del buf[:bytes_unread - 2]

                    if data:
                        yield data
                        bytes_unread -= len(data)

                    paused = False
                    continue

                if bytes_unread == 2:
                    if len(buf) < 2:
                        paused = False
                        continue

                    if not buf.startswith(b'\r\n'):
                        raise BadRequest('bad chunked encoding: '
                                         'invalid chunk terminator')

                    if chunk_size == 0:
                        if self.server.options['experimental']:
                            self.server.queue[0].queue.appendleft(buf[2:])
                        else:
                            self.http_keepalive = False

                        del buf[:]
                        break

                # bytes_unread should only be either 0 or 2 at this point
                i = buf.find(b'\r\n', bytes_unread)

                if i == -1:
                    if len(buf) > 64:
                        raise BadRequest('bad chunked encoding: no chunk size')

                    paused = False
                    continue

                try:
                    chunk_size = parse_int(
                        buf[bytes_unread:i].split(b';', 1)[0], 16
                    )
                except ValueError as exc:
                    raise BadRequest('bad chunked encoding') from exc

                data = bytes(buf[i + 2:i + 2 + chunk_size])
                del buf[:i + 2 + chunk_size]

                if data:
                    yield data

                bytes_unread = chunk_size - len(data) + 2
                paused = True

        self.content_length = 0

        if self._stream is not None:  # extra for recv(size=-1)
            yield b''

    @property
    def params(self):
        return self.context

    @property
    def query(self):
        try:
            return self.params['query']
        except KeyError:
            self.params['query'] = {}

            if self.query_string != b'':
                self.params['query'] = parse_qs(
                    self.query_string.decode('latin-1'), max_num_fields=100
                )

            return self.params['query']

    @property
    def cookies(self):
        try:
            return self.params['cookies']
        except KeyError:
            self.params['cookies'] = {}

            if b'cookie' in self.headers:
                cookie = b';'.join(self.headers[b'cookie'])

                for k, v in parse_fields(cookie):
                    k = k.decode('latin-1')

                    if k in self.params['cookies']:
                        self.params['cookies'][k].append(v.decode('latin-1'))
                    else:
                        self.params['cookies'][k] = [v.decode('latin-1')]

            return self.params['cookies']

    async def form(self, max_fields=100, *, max_size=8 * 1048576):
        try:
            return self.params['post']
        except KeyError:
            content_type = self.headers.getlist(b'content-type', b';')
            self.params['post'] = {}
            self.params['files'] = {}

            if b'application/x-www-form-urlencoded' in content_type:
                async for data in self.stream():
                    self._body.extend(data)

                    if self.body_size > max_size:
                        raise ValueError('form size limit reached') from None

                if 2 < self.body_size <= max_size:
                    self.params['post'] = parse_qs(
                        self._body.decode('latin-1'),
                        max_num_fields=max_fields
                    )

                return self.params['post']

            if b'multipart/form-data' in content_type:
                self._files = self.files(max_fields, max_file_size=max_size)

                async for part in self._files:
                    if not part.pop('eof'):
                        raise ValueError(
                            'fragmented file. consider increasing the '
                            'max_size limit or stream using request.files()'
                        ) from None

                    name = part.pop('name', '')

                    if 'filename' in part:
                        if part['filename'] == '':
                            continue

                        if name in self.params['files']:
                            self.params['files'][name].append(part)
                        else:
                            self.params['files'][name] = [part]
                    else:
                        if name in self.params['post']:
                            self.params['post'][name].append(
                                part['data'].decode()
                            )
                        else:
                            self.params['post'][name] = [part['data'].decode()]

                return self.params['post']

            raise BadRequest('invalid Content-Type') from None

    async def files(self, max_files=1024, *, max_file_size=100 * 1048576):
        if self.eof():
            return

        if self._files is None:
            self._files = self.files(max_files, max_file_size=max_file_size)

            async for part in self._files:
                yield part

            return

        for key, boundary in parse_fields(self.content_type):
            if key == b'boundary' and boundary:
                break
        else:
            raise BadRequest('missing boundary')

        boundary_size = len(boundary)
        body = bytearray()

        header_size = 0
        body_size = 0
        paused = False
        part = None  # represents a field/file received in a multipart request

        while max_files > 0:
            data = b''

            if not paused:
                data = await self.read()

                if not data:
                    if self._read_buf.startswith(b'--%s--' % boundary):
                        del self._read_buf[:]  # set eof()
                        return

                    if header_size == 1 or body_size == -1:
                        raise BadRequest(
                            'malformed multipart/form-data: incomplete read'
                        )

            if part is None:
                self._read_buf.extend(data)
                header_size = self._read_buf.find(b'\r\n\r\n') + 2

                if header_size == 1:
                    if len(self._read_buf) > 8192:
                        raise BadRequest(
                            'malformed multipart/form-data: header too large'
                        )

                    paused = False
                else:
                    body.extend(self._read_buf[header_size + 2:])
                    part = MultipartFile(self._files)

                    # use find() instead of startswith() to ignore the preamble
                    if self._read_buf.find(b'--%s\r\n' % boundary,
                                           0, header_size) != -1:
                        header = self.header.parse(self._read_buf,
                                                   header_size).headers

                        if b'content-disposition' in header:
                            for k, v in parse_fields(
                                    header[b'content-disposition'][0]):
                                part[k.decode('latin-1')] = v.decode('latin-1')

                        if b'content-type' in header:
                            part['type'] = header[b'content-type'][
                                                  0].decode('latin-1')
                continue

            body.extend(data)
            body_size = body.find(b'\r\n--%s' % boundary)

            if body_size == -1:
                if len(body) >= max_file_size > boundary_size + 4:
                    part['data'] = bytes(body[:-boundary_size - 4])
                    part['eof'] = False
                    yield part

                    del body[:-boundary_size - 4]

                paused = False
                continue

            part['data'] = bytes(body[:body_size])
            part['eof'] = True
            yield part

            self._read_buf[:] = body[body_size + 2:]
            paused = True
            part = None
            max_files -= 1

            del body[:]



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import base64
import hashlib
import os

from .http_exceptions import WebSocketClientClosed, WebSocketServerClosed

_MAGIC = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'


class WebSocket:
    def __init__(self, request, response):
        self.request = request
        self.response = response
        self.max_payload_size = request.server.options['ws_max_payload_size']
        self.ping_interval = request.server.options['request_timeout'] / 2

        self.fin = 1
        self.opcode = 0

    def __aiter__(self):
        return self

    async def __anext__(self):
        if not self.request.upgraded:
            await self.accept()

        try:
            return await self.receive()
        except WebSocketClientClosed as exc:
            await self.close()
            raise StopAsyncIteration from exc

    async def accept(self):
        sha1_hash = hashlib.sha1(  # nosec B303,B324
            self.request.headers[b'sec-websocket-key'][0] + _MAGIC
        ).digest()
        accept_key = base64.b64encode(sha1_hash)

        self.response.set_status(101, b'Switching Protocols')
        self.response.set_header(b'Upgrade', b'websocket')
        self.response.set_header(b'Sec-WebSocket-Accept', accept_key)
        await self.response.write()

    async def recv(self, timeout=None):
        try:
            b1, b2 = await self.request.recv(2, timeout=timeout)
        except ValueError as exc:
            raise WebSocketClientClosed(
                'connection closed: recv failed'
            ) from exc

        fin = (b1 & 0x80) >> 7
        opcode = b1 & 0x0f
        is_masked = (b2 & 0x80) >> 7
        payload_length = b2 & 0x7f

        if opcode != 0:
            if self.fin != 1:
                raise WebSocketServerClosed('unexpected start', code=1002)

            if fin == 0:  # start of fragmented message
                self.fin = 0

            self.opcode = opcode
        else:  # continuation frame
            if self.fin == 1:
                raise WebSocketServerClosed('unexpected continuation',
                                            code=1002)

            if fin == 1:  # end of fragmented message, reset
                self.fin = 1

        if payload_length == 126:
            payload_length = int.from_bytes(await self.request.recv(2),
                                            byteorder='big')
        elif payload_length == 127:
            payload_length = int.from_bytes(await self.request.recv(8),
                                            byteorder='big')

            if payload_length > self.max_payload_size:
                raise WebSocketServerClosed(
                    '%d exceeds maximum payload size (%d)' %
                    (payload_length, self.max_payload_size),
                    code=1009
                )

        if is_masked:
            masking_key = await self.request.recv(4)
            payload_data = await self.request.recv(payload_length)
            unmasked_data = bytes(
                data_byte ^ masking_key[i % 4] for i, data_byte in enumerate(
                    payload_data)
            )
        else:
            unmasked_data = await self.request.recv(payload_length)

        if opcode == 1 or opcode == 2 or opcode == 0:
            return unmasked_data

        # ping
        if opcode == 9 and payload_length < 126:
            await self.pong(unmasked_data)
            return b''

        # pong
        if opcode == 10 and payload_length < 126:
            return b''

        if opcode == 8:
            code = 1005

            if unmasked_data != b'':
                code = int.from_bytes(unmasked_data[:2], byteorder='big')

            raise WebSocketClientClosed(
                'connection closed (%d)' % code,
                code=code
            )

        raise WebSocketServerClosed(
            'unsupported opcode %x with payload length %d' %
            (opcode, payload_length),
            code=1008
        )

    async def receive(self, timeout=None):
        payload = bytearray()

        while self.opcode != 8:
            loop = self.request.server.loop  # may raise RuntimeError
            coro = self.ping()
            timer = loop.call_at(loop.time() + self.ping_interval,
                                 self.request.server.create_task, coro)

            try:
                frame = await self.recv(timeout)
            except TimeoutError as exc:
                raise WebSocketServerClosed('receive timeout',
                                            code=1000) from exc
            finally:
                timer.cancel()
                coro.close()

            if frame == b'':
                # got empty bytes (pong). continue and keep pinging
                continue

            payload.extend(frame)

            if len(payload) > self.max_payload_size:
                raise WebSocketServerClosed('maximum payload size exceeded',
                                            code=1009)

            if self.fin == 1:
                break
        else:
            raise WebSocketClientClosed('connection already closed')

        if self.opcode == 1:
            return payload.decode('utf-8')

        return payload

    @staticmethod
    def create_frame(payload_data, fin=1, opcode=None, mask=False):
        if opcode is None:
            if isinstance(payload_data, str):
                opcode = 1
            else:
                opcode = 2

        if opcode == 1:
            payload_data = payload_data.encode('utf-8')

        b1 = (fin << 7) | opcode
        payload_length = len(payload_data)

        if payload_length < 126:
            b2 = payload_length
            payload_length_data = b''
        elif payload_length < 65536:
            b2 = 126
            payload_length_data = payload_length.to_bytes(2, byteorder='big')
        else:
            b2 = 127
            payload_length_data = payload_length.to_bytes(8, byteorder='big')

        if mask:
            b2 |= (1 << 7)
            masking_key = os.urandom(4)

        frame_header = bytes([b1, b2]) + payload_length_data

        if mask:
            masked_payload_data = bytes(
                data_byte ^ masking_key[i % 4] for i, data_byte in enumerate(
                    payload_data)
            )
            payload_data = masking_key + masked_payload_data

        return frame_header + payload_data

    async def send(self, payload_data, fin=1, opcode=None):
        await self.response.send(
            __class__.create_frame(payload_data, fin=fin, opcode=opcode)
        )

    async def ping(self, data=b''):
        # ping only if this connection is still listed,
        # otherwise let the recv timeout drop it
        if self.request.server in self.request.server.globals.connections:
            await self.send(data, opcode=9)

    async def pong(self, data=b''):
        await self.send(data, opcode=10)

    async def close(self, code=1000, *, timeout=None):
        try:
            await self.send(code.to_bytes(2, byteorder='big'), opcode=8)
            await self.receive(timeout)
        except RuntimeError:
            return
        except (WebSocketClientClosed, WebSocketServerClosed):
            pass

        self.response.close()



# SPDX-License-Identifier: MIT
# Copyright (c) 2023 Anggit Arfanto

import asyncio

from .exceptions import LifespanError, LifespanProtocolUnsupported


class ASGILifespan:
    def __init__(self, app, options):
        self._loop = app.loop
        self._logger = app.logger
        self._options = options

        self._queue = asyncio.Queue()
        self._waiter = self._loop.create_future()
        self._task = self._loop.create_task(self.main(options['app']))

    async def main(self, app):
        try:
            scope = {
                'type': 'lifespan',
                'asgi': {'version': '3.0', 'spec_version': '2.0'},
                'state': self._options['state']
            }

            await app(scope, self.receive, self.send)
        finally:
            self._waiter.cancel()

    def startup(self):
        self._queue.put_nowait({'type': 'lifespan.startup'})
        self._logger.info('lifespan: startup')

    def shutdown(self):
        self._queue.put_nowait({'type': 'lifespan.shutdown'})
        self._logger.info('lifespan: shutdown')

    async def receive(self):
        data = await self._queue.get()
        self._queue.task_done()

        return data

    async def send(self, data):
        if data['type'] in ('lifespan.startup.complete',
                            'lifespan.shutdown.complete'):
            self._waiter.set_result(self._loop.create_future())
            self._logger.info(data['type'])
        elif data['type'] in ('lifespan.startup.failed',
                              'lifespan.shutdown.failed'):
            raise LifespanError(
                '%s: %s' % (data['type'], data.get('message', ''))
            )
        else:
            raise LifespanProtocolUnsupported

    async def exception(self, timeout=30):
        timer = self._loop.call_at(self._loop.time() + timeout,
                                   self._waiter.cancel)

        try:
            self._waiter = await self._waiter
        except asyncio.CancelledError:
            if not self._task.done():
                self._task.cancel()
                self._logger.error('lifespan: timeout after %gs', timeout)
                raise

            exc = self._task.exception()

            if exc:
                if isinstance(exc, LifespanError):
                    return exc

                if isinstance(exc, LifespanProtocolUnsupported):
                    self._logger.info(exc)
                else:
                    self._logger.info(
                        '%s: %s',
                        LifespanProtocolUnsupported.message,
                        str(exc) or exc.__class__.__name__
                    )
        finally:
            timer.cancel()
